# Copyright © The University of Edinburgh, 2022.
# Development has been supported by GSK.

from .tcga import TCGA
from .cmap import CMAP
from .iris import Iris
from .iris import Iris_2_views
from .breast_cancer import BreastCancer


